import json
from contextlib import contextmanager

from sqlalchemy import Column, Text, String, Integer, create_engine, Boolean
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import sessionmaker

from jarbas_hive_mind.database import Base
from jarbas_hive_mind.settings import CLIENTS_DB


@contextmanager
def session_scope(db):
    """Provide a transactional scope around a series of operations."""
    Session = sessionmaker(bind=db)
    session = Session()
    try:
        yield session
        session.commit()
    except:
        session.rollback()
        raise
    finally:
        session.close()


class Client(Base):
    __tablename__ = "clients"
    id = Column(Integer, primary_key=True)
    description = Column(Text)
    api_key = Column(String)
    name = Column(String)
    mail = Column(String)
    last_seen = Column(Integer, default=0)
    is_admin = Column(Boolean, default=False)
    blacklist = Column(Text)  # json string


class ClientDatabase(object):
    def __init__(self, path=CLIENTS_DB, debug=False):
        self.db = create_engine(path)
        self.db.echo = debug

        Base.metadata.create_all(self.db)

    def update_timestamp(self, key, timestamp):
        with session_scope(self.db) as session:
            user = self.get_client_by_api_key(key)
            if not user:
                return False
            user.last_seen = timestamp
            return True

    def delete_client(self, key):
        with session_scope(self.db) as session:
            user = self.get_client_by_api_key(key)
            if user:
                session.delete(user)
                return True
            return False

    def change_api(self, old_key, new_key):
        with session_scope(self.db) as session:
            user = self.get_client_by_api_key(old_key)
            if not user:
                return False
            user.api_key = new_key
        return True

    def change_name(self, new_name, key):
        with session_scope(self.db) as session:
            user = self.get_client_by_api_key(key)
            if not user:
                return False
            user.name = new_name
        return True

    def change_blacklist(self, blacklist, key):
        if isinstance(blacklist, dict):
            blacklist = json.dumps(blacklist)
        with session_scope(self.db) as session:
            user = self.get_client_by_api_key(key)
            if not user:
                return False
            user.blacklist = blacklist
        return True

    def get_blacklist_by_api_key(self, api_key):
        with session_scope(self.db) as session:
            user = session.query(Client).filter_by(api_key=api_key).first()
            return json.loads(user.blacklist)

    def get_client_by_api_key(self, api_key):
        with session_scope(self.db) as session:
            return session.query(Client).filter_by(api_key=api_key).first()

    def get_client_by_name(self, name):
        with session_scope(self.db) as session:
            return session.query(Client).filter_by(name=name).all()

    def add_client(self, name=None, mail=None, key="", admin=False,
                   blacklist="{}"):
        if isinstance(blacklist, dict):
            blacklist = json.dumps(blacklist)

        with session_scope(self.db) as session:
            user = self.get_client_by_api_key(key)
            if user:
                user.name = name
                user.mail = mail
                user.blacklist = blacklist
                user.is_admin = admin
            else:
                user = Client(api_key=key, name=name, mail=mail,
                              blacklist=blacklist, id=self.total_clients() + 1,
                              is_admin=admin)
                session.add(user)

    def total_clients(self):
        with session_scope(self.db) as session:
            return session.query(Client).count()

    def commit(self, handler):
        Session = sessionmaker(bind=self.db)
        session = Session()
        try:
            handler(session)
            session.commit()
        except IntegrityError:
            session.rollback()
            raise
        finally:
            session.close()




from jarbas_hive_mind.terminals.webchat import platform, logger, JarbasWebChatTerminal
import base64
from twisted.internet import reactor, ssl


def connect_to_hivemind(host="127.0.0.1",
                        port=5678, name="Jarbas WebChat Terminal",
                        key="webchat_key", useragent=platform):
    authorization = bytes(name + ":" + key, encoding="utf-8")
    usernamePasswordDecoded = authorization
    api = base64.b64encode(usernamePasswordDecoded)
    headers = {'authorization': api}
    address = u"wss://" + host + u":" + str(port)
    logger.info("[INFO] connecting to hive mind at " + address)
    terminal = JarbasWebChatTerminal(address, headers=headers,
                                     useragent=useragent)
    contextFactory = ssl.ClientContextFactory()
    reactor.connectSSL(host, port, terminal, contextFactory)
    reactor.run()


if __name__ == '__main__':
    # TODO parse args
    connect_to_hivemind()



import json
import ssl
import time
from functools import wraps

from flask import Flask, make_response
from flask import request, Response
from flask_sslify import SSLify

from jarbas_hive_mind.database.client import ClientDatabase
from jarbas_hive_mind.nodes import gen_key
from jarbas_hive_mind.settings import DEFAULT_SSL_CRT, DEFAULT_SSL_KEY

users = ClientDatabase()


def nice_json(arg):
    response = make_response(json.dumps(arg, sort_keys=True, indent=4))
    response.headers['Content-type'] = "application/json"
    return response


def add_response_headers(headers=None):
    """This decorator adds the headers passed in to the response"""
    headers = headers or {}

    def decorator(f):
        @wraps(f)
        def decorated_function(*args, **kwargs):
            resp = make_response(f(*args, **kwargs))
            h = resp.headers
            for header, value in list(headers.items()):
                h[header] = value
            return resp

        return decorated_function

    return decorator


def noindex(f):
    """This decorator passes X-Robots-Tag: noindex"""
    return add_response_headers({'X-Robots-Tag': 'noindex'})(f)


def donation(f):
    """This decorator passes btc request """
    return add_response_headers({'BTC': '1aeuaAijzwK4Jk2ixomRkqjF6Q3JxXp9Q',
                                 "Patreon": "patreon.com/jarbasAI",
                                 "Paypal": "paypal.me/jarbasAI"})(
        f)


def check_auth(key):
    """This function is called to check if a api key is valid."""
    user = users.get_client_by_api_key(key)
    if not user:
        return False
    users.update_timestamp(key, time.time())
    return True


def check_admin_auth(key):
    """This function is called to check if a admin api key is valid."""

    user = users.get_client_by_api_key(key)
    if not user:
        return False
    if not user.is_admin:
        return False
    users.update_timestamp(key, time.time())
    return True


def authenticate():
    """Sends a 401 response that enables basic auth"""
    return Response(
        'Could not verify your access level for that URL.\n'
        'You have to login with proper credentials', 401,
        {'WWW-Authenticate': 'Basic realm="Api Key Required"'})


def requires_auth(f):
    @wraps(f)
    def decorated(*args, **kwargs):
        auth = request.headers.get('Authorization', '')
        if not auth or not check_auth(auth):
            return authenticate()
        return f(*args, **kwargs)

    return decorated


def requires_admin(f):
    @wraps(f)
    def decorated(*args, **kwargs):
        auth = request.headers.get('Authorization', '')
        if not auth or not check_admin_auth(auth):
            print("not admin")
            return authenticate()
        return f(*args, **kwargs)

    return decorated


def get_app():
    app = Flask(__name__)
    sslify = SSLify(app)

    @app.route("/", methods=['GET'])
    @noindex
    @donation
    def hello():
        return nice_json({
            "uri": "/",
            "welcome to jarbas microservices": {
                "under": "construction"
            }
        })

    @app.route("/revoke_key/<key>", methods=['PUT'])
    @noindex
    @donation
    @requires_admin
    def revoke_key(key):
        if users.delete_client(key):
            result = {"removed": True}
        else:
            result = {"removed": False, "error": "does not exist"}
        return nice_json(
            result
        )

    @app.route("/new_user/<key>/<mail>/<name>", methods=['PUT'])
    @noindex
    @donation
    @requires_admin
    def add_user(key, mail, name):
        result = {"success": False}
        if users.add_client(name, mail, key):
            result = {"success": True}
        return nice_json(
            result
        )

    @app.route("/get_key", methods=['GET'])
    @noindex
    @donation
    @requires_admin
    def new_key():
        key = gen_key(save=False)
        return nice_json(
            {"key": key}
        )

    return app


def start(app, port=6666):
    cert = DEFAULT_SSL_CRT
    key = DEFAULT_SSL_KEY
    context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
    context.load_cert_chain(cert, key)
    app.run(host="0.0.0.0", port=port, debug=False, ssl_context=context)


if __name__ == "__main__":
    start(get_app())
